/*
 * This is SSSim: the Simple & Stupid Simulator
 *
 *  Copyright (c) 2008 Luca Abeni
 *  Copyright (c) 2008 Csaba Kiraly
 *
 *  This is free software; see gpl-3.0.txt
 */

#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "config.h"
#include "core.h"
#include "graph.h"

//#define neighbourhood_generate_src neighbourhood_generate_hypercube
#define neighbourhood_generate_src neighbourhood_generate_uni

//#define graph_generate create_adj_matrix
#define graph_generate graph_generate_stupid

static void shuffle_int(int dimesion, int *input)
{
  int i, dummy;

  for (i=0; i < dimesion-1; i++){
    int64_t rnd_int;

    rnd_int  = rand();
    rnd_int = (rnd_int * (dimesion - i)) / RAND_MAX;
    //rnd_int = (int)uniform(0.0, (double)(dimesion-i), seed);
    dummy = input[i];
    input[i] = input[i+rnd_int]; 
    input[i+rnd_int] = dummy;
  }
}

static struct peer **graph_generate_stupid(int size, struct peer *peers, int np)
{
  int n;
  struct peer **graph;
  int edges[np];

  graph = malloc(sizeof(struct peer *) * size * np);
  memset(graph, 0, sizeof(struct peer *) * size * np);
  memset(edges, 0, sizeof(edges));
  for (n = 0; n < np; n++) {
    int j, cnt;

    cnt = 0;
    for (j = np - 1; j > n; j--) {
      if (edges[j] < size) {
        cnt++;
      }
    }
    if (size - edges[n] > cnt) {
      fprintf(stderr, "Generation failed at %d\n", n);
      free(graph);
      return NULL;
    }
    while (edges[n] < size) {
      int64_t target;

      target = rand();
      target = (target * np) / RAND_MAX;
      if (target == n) target = -1;
      for (j = 0; j < edges[n]; j++) {
        if (graph[n * size + j] == &peers[target]) {
          target = -1;
        }
      }
      if (edges[target] == size) {
        target = -1;
      }
      if (target >= 0) {
        graph[n * size + edges[n]] = &peers[target];
        graph[target * size + edges[target]] = &peers[n];
        edges[n]++;
        edges[target]++;
      }
    }
  }

  return graph;
}

static struct peer **graph_generate_hypercube(int size, struct peer *peers, int np)
{
  int n,i,target;
  struct peer **graph;

  int log2n = ceil(log2((double)np));
  if (log2n != size) { fprintf(stderr, "Error: size should be ceil(log2(n))\n"); exit(2); }

  graph = malloc(sizeof(struct peer *) * size * np);
  memset(graph, 0, sizeof(struct peer *) * size * np);
  for (n = 0; n < np; n++) {
    //n = (p->id == -1) ? 0 : p->id;
    for (i = 0; i<log2n; i++) {
      target = n ^ (1<<i);
      if (target < np){
        graph[n * size + i] = &peers[target];	//bidir is guaranteed by XOR
        fprintf(stdout, "edge %d -> %d\n",n,target);
      } else {
        fprintf(stderr, "Warning: edge can't be created\n");
      }
    }
  }

  return graph;
}

struct peer **neighbourhood_generate_hypercube(int size, struct peer *peers, int np, int id)
{
  int base,i,target;
  struct peer **neigh;

  int log2n = ceil(log2((double)np));
  if (log2n != size) { fprintf(stderr, "Error: size should be ceil(log2(n))\n"); exit(2); }

  neigh = malloc(sizeof(struct peer *) * size);
  memset(neigh, 0, sizeof(struct peer *) * size);

  base = (id == -1) ? 0 : id;
  for (i = 0; i<log2n; i++) {
    target = base ^ (1<<i);
    if (target < np){
      neigh[i] = &peers[target];	//bidir is guaranteed by XOR
      fprintf(stdout, "edge %d -> %d\n",id,target);
    } else {
      fprintf(stderr, "Warning: edge can't be created\n");
    }
  }
  return neigh;
}

static struct peer **create_adj_matrix(int num_neighbors, struct peer *peers, int num_nodes)
{
  struct peer **input_adj_matrix;
  int node_i, neighbor_k, *actual_neighbors, *reference_vector;

 /* allocate variuos structures:
    - the vector that memorize the number of neighbors for each node
    - the vector that is shuffled */
  actual_neighbors = calloc(num_nodes, sizeof(int));
  reference_vector = calloc(num_nodes, sizeof(int));
  input_adj_matrix = malloc(sizeof(struct peer *) * num_nodes * num_neighbors);

fprintf(stderr, "Generating %d %p %d\n", num_nodes, peers, num_neighbors);
  /* analyze each node */
  for (node_i=0; node_i < num_nodes; node_i++) {
     neighbor_k = actual_neighbors[node_i];
  
fprintf(stderr, "N-K: %d\n", neighbor_k);
     if ((num_neighbors - neighbor_k > 0) && 
         (num_nodes - node_i > num_neighbors - neighbor_k)){
        int i;
 
        /* populate the reference vector and shuffle it*/
        reference_vector[0] = 0;
        for (i=1; i <= num_neighbors - neighbor_k; i++) {
 fprintf(stderr, "Setting %d to 1\n", i);
            reference_vector[i] = 1;
}
        for (i=num_neighbors - neighbor_k + 1; i < num_nodes - node_i; i++) {
 fprintf(stderr, "Setting %d to 0\n", i);
            reference_vector[i] = 0;
}
 
        shuffle_int(num_nodes - node_i -1, &reference_vector[1]);

for (i=0; i < num_nodes - node_i; i++) fprintf(stderr, "%d ", reference_vector[i]);
fprintf(stderr, "\n");
        /* populate the adjacency matrix */
        for (i=1; i < num_nodes - node_i; i++) {
fprintf(stderr, "%d-th (over %d) neighbour\n", i, num_nodes - node_i);
            if ((reference_vector[i] == 1)&&(actual_neighbors[i + node_i]<num_neighbors)) {
               input_adj_matrix[node_i * num_neighbors + actual_neighbors[node_i]] = &peers[i + node_i];
fprintf(stderr, "%d: %d <---> %d\n", node_i * num_neighbors + actual_neighbors[node_i], node_i, i + node_i);
               actual_neighbors[node_i]++;
        
               input_adj_matrix[(i + node_i) * num_neighbors + actual_neighbors[i + node_i]] = &peers[node_i];
               actual_neighbors[i + node_i]++;
            }//if
        }//for
    }//if
  }//for
 
  /* space deallocation */   
//  free(actual_neighbors);
//  free(reference_vector);

  return input_adj_matrix;
}


struct peer **neighbourhood_generate_uni(int size, struct peer *peers, int n, int id)
{
  int i;
  struct peer **neigh;

  neigh = malloc(sizeof(struct peer *) * size);
  for (i = 0; i < size;) {
    int64_t target;
    int j;

    target = rand();
    target = (target * n) / RAND_MAX;
    if (target == id) target = -1;
    for (j = 0; j < i; j++) {
      if (neigh[j] == &peers[target]) {
        target = -1;
      }
    }
    if (target >= 0) {
      neigh[i] = &peers[target];
      i++;
    }
  }

  return neigh;
}

struct peer **neighbourhood_generate_fake(int size, struct peer *peers, int n, int id)
{
  int i;
  static struct peer **p;
 
  if (n != size) {
    return NULL;
  }

  if (p == NULL) {
    p = malloc(sizeof(struct peer *) *size);
    for (i = 0; i < size; i++ ) {
      p[i] = &peers[i];
    }
  }

  return p;
}

struct peer **neighbourhood_generate_bidir(int size, struct peer *peers, int np, int id)
{
  static struct peer **graph;

  if (id < 0) {
    return neighbourhood_generate_src(size, peers, np, id);
  }

  if (np == size) {
    return neighbourhood_generate_fake(size, peers, np, id);
  }

  while (graph == NULL) {
    graph = graph_generate(size, peers, np);
  }
  return graph + id * size;
}

struct peer **neighbourhood_generate_pipe(int size, struct peer *peers, int n, int id)
{
  struct peer **p;
 
  if (size != 1) {
    return NULL;
  } 
  p = malloc(sizeof(struct peer *) *size);
  p[0] = &peers[id + 1];
  if (p[0] == &peers[n]) p[0] = &peers[id];


  return p;
}

void neighbourhood_print(struct peer **n, int size)
{
  int i;

  for (i = 0; i < size; i++) {
    printf("%d ", n[i]->id);
  }
}

void graph_dotprint(FILE *f, struct peer *s, struct peer *n, int np)
{
  int i, j;

  fprintf(f, "strict graph overlay {\n");
  if (s) {
    for (j = 0; j < s->neigh_size; j++) {
      fprintf(f, "  S -- P%d;\n", s->neighbour[j]->id);
    }
  }
  for (i = 0; i < np; i++) {
    for (j = 0; j < n[i].neigh_size; j++) {
      fprintf(f, "  P%d -- P%d;\n", n[i].id, n[i].neighbour[j]->id);
    }
  }
  fprintf(f, "}\n");
}
